import csv
import time
import random
import logging
from typing import Optional

import gevent
from pymodbus.client.sync import ModbusTcpClient
from pymodbus.transaction import ModbusRtuFramer, ModbusSocketFramer

from .base import TerrawareDevice


class ModbusDevice(TerrawareDevice):

    def __init__(self, controller, server_path, host, port, settings, polling_interval, diagnostic_mode):
        settings_items = settings.split(';')
        self._controller = controller
        self._server_path = server_path
        self._polling_interval = polling_interval
        self._host = host
        self._unit = 1  # aka modbus slave number
        for setting in settings_items:
            if setting.startswith('unit='):
                self._unit = int(setting.split('=')[1])
        print('initializing device %s (%s:%d, unit: %d)' % (server_path, host, port, self._unit))
        self.last_update_time = None
        self._diagnostic_mode = diagnostic_mode
        framer = ModbusRtuFramer if ('rtu-over-tcp' in settings_items) else ModbusSocketFramer
        self._modbus_client = ModbusTcpClient(host, port=port, framer=framer)
        self._read_holding = ('holding' in settings_items)
        self._seq_infos = []

        # load seqeuence info
        with open('config/%s.csv' % server_path) as csvfile:
            lines = csv.DictReader(csvfile)
            for line in lines:
                self._seq_infos.append(line)

    def server_path(self):
        return self._server_path

    def run(self):
        logging.info('starting modbus monitoring/control for %s; polling interval: %.1fs', self._server_path, self._polling_interval)
        if self._host != 'sim':
            self._modbus_client.connect()
        while True:
            if self._host != 'sim' and not self._modbus_client.is_socket_open():
                self._modbus_client.connect()
            seq_values = {}
            for seq_info in self._seq_infos:
                address = int(seq_info['address'], 0)
                value = self.read_register(address, seq_info['type'], self._unit)
                if value is not None:
                    value *= float(seq_info['scale_factor'])
                    seq_rel_path = self._server_path + '/' + seq_info['name']
                    full_seq_name = self._controller.path_on_server() + '/' + seq_rel_path
                    if self._diagnostic_mode:
                        print('    %s: %.2f' % (seq_rel_path, value))
                    self._controller.sequences.update_value(seq_rel_path, value)
                    if int(seq_info['send_to_server']):
                        self._controller.sequence.create(seq_rel_path, 'numeric', decimal_places=2)
                        seq_values[full_seq_name] = value
            if seq_values:
                self._controller.sequences.update_multiple(seq_values)
                self.last_update_time = time.time()
            print('received %d value(s) from %s' % (len(seq_values), self._server_path))
            gevent.sleep(self._polling_interval)

    def read_register(self, address, register_type, unit):
        if self._host == 'sim':
            return random.randint(1, 100)
        if register_type.endswith('32'):
            count = 2
        else:
            count = 1
        if self._read_holding:
            result = self._modbus_client.read_holding_registers(address, count, unit=unit)
        else:
            result = self._modbus_client.read_input_registers(address, count, unit=unit)
        if not hasattr(result, 'registers'):
            return None
        if register_type == 'uint16':
            return result.registers[0]
        elif register_type == 'sint16':
            v = result.registers[0]
            if v > 0x7fff:  # rough sign manipulation; should check/fix
                v = v - 0x10000
            return v
        elif register_type == 'uint32':
            return result.registers[0] * 0x10000 + result.registers[1]
        elif register_type == 'sint32':
            v = result.registers[0] * 0x10000 + result.registers[1]
            if v > 0x7fffffff:  # rough sign manipulation; should check/fix
                v = v - 0x100000000
            return v



import time
import logging
from xml.etree import ElementTree

try:
    from httplib import HTTPConnection
except ModuleNotFoundError:
    from http.client import HTTPConnection
import gevent

from .base import TerrawareDevice


class RelayDevice(TerrawareDevice):

    def __init__(self, controller, server_path, host, port, settings, polling_interval, diagnostic_mode):
        print('initializing device %s (%s:%d)' % (server_path, host, port))
        self._controller = controller
        self._server_path = server_path
        self._host = host
        self._port = port
        self._polling_interval = polling_interval
        self._last_update_time = None
        self._sim_state = 0

    def server_path(self):
        return self._server_path

    def run(self):
        logging.info('starting relay monitoring/control for %s; polling interval: %.1fs', self._server_path, self._polling_interval)
        while True:
            print('polling %s now' % self._server_path)
            v = self.read_state()
            seq_rel_path = self._server_path + '/relay-1'
            if False:
                print('    %s: %d' % (seq_rel_path, v))
            self._controller.sequences.update(seq_rel_path, v)
            self._controller.sequences.update_value(seq_rel_path, v)
            self._last_update_time = time.time()
            gevent.sleep(self._polling_interval)

    def read_state(self):
        if self._host == 'sim':
            xml = self.sample_data()
        else:
            conn = HTTPConnection('%s:%d' % (self._host, self._port))
            conn.request('GET', '/state.xml')
            xml = conn.getresponse().read()
            conn.close()
        tree = ElementTree.fromstring(xml)
        return int(tree.find('relay1state').text)

    def set_state(self, state):
        if self._host == 'sim':
            self._sim_state = state
        else:
            conn = HTTPConnection('%s:%d' % (self._host, self._port))
            conn.request('GET', '/state.xml?relay1state=%d' % state)
            conn.close()

    def sample_data(self):
        return f'''<?xml version='1.0' encoding='utf-8'?>
            <datavalues>
                <relay1state>{self._sim_state}</relay1state>
                <relay2state>0</relay2state>
                <relay3state>0</relay3state>
                <relay4state>0</relay4state>
            </datavalues>'''



from bluepy import btle


# returns a list of currently accessible Blue Maestro devices; each device is returned as a dictionary of information
def find_blue_maestro_devices(iface=0, timeout=2, verbose=False):
    btle.Debugging = verbose
    scanner = btle.Scanner(iface)
    devices = scanner.scan(timeout)
    dev_infos = []
    for d in devices:
        manufacturer = d.getValueText(255)
        if manufacturer and manufacturer.startswith('3301'):
            label = d.getValueText(9).decode()
            if label and len(d.rawData) > 17:
                rssi = d.rssi
                vals = [ord(v) for v in d.rawData]
                dev_info = {
                    'label': label,
                    'rssi': d.rssi,
                    'temperature': (vals[13] * 256 + vals[14]) * 0.1,
                    'humidity': (vals[15] * 256 + vals[16]) * 0.1,
                }
                dev_infos.append(dev_info)
    return dev_infos


if __name__ == "__main__":
    dev_infos = find_blue_maestro_devices()
    for d in dev_infos:
        print('label: %s, rssi: %d, t: %.1f, h: %.1f' % (d['label'], d['rssi'], d['temperature'], d['humidity']))



from abc import ABC, abstractmethod
from typing import Optional


class TerrawareDevice(ABC):
    """Base class for device implementations."""

    last_update_time: Optional[float] = None
    """What time the device was last updated."""

    @abstractmethod
    def run(self) -> None:
        """Main loop for this device.

        This method is run in a greenlet and should not return. It should use gevent to perform blocking operations.
        """
        ...

    @abstractmethod
    def server_path(self) -> str:
        """Return the server-side path of this device."""
        ...



import gevent
from gevent import monkey
monkey.patch_all()

# standard library imports
import csv
import time
import gevent
import logging
from typing import List

# other imports
from rhizo.controller import Controller
from .base import TerrawareDevice
from .relay_device import RelayDevice
from .modbus_device import ModbusDevice


# manages a set of devices; each device handles a connection to physical hardware
class DeviceManager(object):

    devices: List[TerrawareDevice]

    def __init__(self):
        self.controller = Controller()
        self.devices = []
        self.start_time = None
        self.diagnostic_mode = self.controller.config.device_diagnostics
        self.handler = None
        if self.controller.config.get('sim', False):
            self.load('config/sim_devices.csv')
        else:
            self.load('config/devices.csv')
        self.controller.messages.add_handler(self)

    # initialize devices using a CSV file
    def load(self, device_list_file_name):
        with open(device_list_file_name) as csvfile:
            reader = csv.DictReader(csvfile)
            for line in reader:
                if int(line['enabled']):
                    device_type = line['type']
                    settings = line['settings']
                    server_path = line['server_path']
                    host = line['host']
                    port = int(line['port'])
                    polling_interval = int(line['polling_interval'])
                    if device_type == 'relay':
                        device = RelayDevice(self.controller, server_path, host, port, settings, polling_interval, self.diagnostic_mode)
                    elif device_type == 'modbus':
                        device = ModbusDevice(self.controller, server_path, host, port, settings, polling_interval, self.diagnostic_mode)
                    else:
                        print('unrecognized device type: %s' % device_type)
                        device = None
                    if device:
                        self.devices.append(device)

    def set_handler(self, handler):
        self.handler = handler

    def handle_message(self, message_type, params):
        if self.handler:
            self.handler.handle_message(message_type, params)

    # launch device polling greenlets
    def run(self):
        for device in self.devices:
            device.greenlet = gevent.spawn(device.run)
        self.start_time = time.time()
        while True:
            try:
                if self.handler:
                    self.handler.update()
            except KeyboardInterrupt:
                print('exiting')
                break
            except Exception as ex:
                logging.warning('exception in handler: %s', ex)
                logging.debug('exception details', ex)
            gevent.sleep(10)

    # find a device by server_path
    def find(self, server_path):
        for device in self.devices:
            if device.server_path() == server_path:
                return device

    # check on devices; restart them as needed; if all is good, send watchdog message to server
    def watchdog_update(self):
        auto_restart = False  # disable auto-restart for now; we seem to occasionally get duplicate device greenlets

        # if it has been a while since startup, start checking device updates
        if time.time() - self.start_time > 30:
            devices_ok = True
            for device in self.devices:
                if device.last_update_time is None or time.time() - device.last_update_time > 10 * 60:
                    logging.info('no recent update for device %s', device.server_path())
                    if auto_restart:
                        device.greenlet.kill()  # this doesn't seem to work; we end up with multiple greenlets for the same device
                        device.greenlet = gevent.spawn(device.run)
                    devices_ok = False

            # if all devices are updating, send a watchdog message to server
            if devices_ok:
                self.controller.send_message('watchdog', {})


if __name__ == '__main__':
    d = DeviceManager()
    d.run



from setuptools import setup
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text()

setup(
    name='terraware-devices',
    version='0.1.3',
    description='Device management for Terraformation hardware systems',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/terraware/terraware-devices',
    author='Terraformation',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        # bluepy is not actively maintained; see https://github.com/IanHarvey/bluepy/issues/403
        # It has a Linux-specific C extension, so skip install on other platforms.
        'bluepy~=1.3; platform_system=="Linux"',
        # gevent versions are date-based so tell us nothing about breaking changes
        'gevent>=20.9',
        # pymodbus bumps its minor version number for each release
        'pymodbus~=2.4',
        # rhizo-server API is under development, so minor versions might have breaking changes
        'rhizo-client~=0.1.1',
    ],
    license='MIT',
    packages=['terraware_devices'],
    python_requires='>=3.7, <4',
)
